unit TreeScan;

{
Description:

Component to help with traversing directory structures
Will notify the owner when file or directory is found.

Note that files are scanned before sub-directories, and that
both file and directory scans are in alphabetical order

Revison history:

1996 Dec 15  V1.0.0  First version
1997 Jan 12  V1.0.2  Add file mask (e.g. *.exe)
1997 Mar 28  V1.0.4  Add event when leaving directory
                     Add BeginUpdate/EndUpdate around list updates
}

interface

uses
  SysUtils, Classes, Forms;              // seems to be the minimum we need

type
  TTreeScanner = class(TComponent)
  private
    { Private declarations }
    FOnFileFound: TNotifyEvent;          // where to notify for file found
    FOnDirectoryFound: TNotifyEvent;     // where to notify for start of directory
    FOnDirectoryDone: TNotifyEvent;      // where to notify for end of directory
    FFoundObject: String;                // full file or directory name
    FInitialDirectory: String;           // starting point in the tree
    FFileMask: String;                   // files to search for e.g. *.EXE
    FSearchInProgress: Boolean;          // true whilst searching
  protected
    { Protected declarations }
    procedure scan_directory (const dir: string);
  public
    { Public declarations }
    Continue: Boolean;                   // caller sets the false to interrupt
    constructor Create (AOwner: TComponent);  override;
    procedure ScanTree;                  // call here to scan tree
  published
    { Published declarations }
    property OnDirectoryFound: TNotifyEvent read FOnDirectoryFound write FOnDirectoryFound;
    property OnDirectoryDone: TNotifyEvent read FOnDirectoryDone write FOnDirectoryDone;
    property OnFileFound: TNotifyEvent read FOnFileFound write FOnFileFound;
    property FileFound: String read FFoundObject;
    property SearchInProgress: Boolean read FSearchInProgress;
    property InitialDirectory: String read FInitialDirectory write FInitialDirectory;
    property FileMask: String read FFileMask write FFileMask;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Davids', [TTreeScanner]);
end;

constructor TTreeScanner.Create (AOwner: TComponent);
begin
  Inherited Create (AOwner);
  FOnDirectoryFound := nil;       // no callback pointer yet for directory start
  FOnDirectoryDone := nil;        // or for directory complete
  FOnFileFound := nil;            // or for found file
  FSearchInProgress := False;     // and we're not yet searching
end;

procedure TTreeScanner.ScanTree;
begin
  FSearchInProgress := True;      // now searching
  FFoundObject := '';             // nothing found as yet

  // Try to get a valid directory string.  If the user hasn't
  // specified a directory, use the current directory...
  if FInitialDirectory = '' then FInitialDirectory := GetCurrentDir;

  // ensure the directory ends in a '\'
  if FInitialDirectory [Length (FInitialDirectory)] <> '\' then
    FInitialDirectory := FInitialDirectory + '\';

  // now remove any file specification components...
  FInitialDirectory := ExtractFileDir (FInitialDirectory);
  // and again ensure that trailing backslash is present....
  if FInitialDirectory [Length (FInitialDirectory)] <> '\' then
    FInitialDirectory := FInitialDirectory + '\';
  // assume we wish to continue (at least to start with!)
  Continue := True;
  // and call the recursive scanning procedure
  scan_directory (FInitialDirectory);
  FSearchInProgress := False;      // note we have finished searching
end;

procedure TTreeScanner.scan_directory (const dir: string);
var
 f: TSearchRec;             // used for scanning a directory
 status: integer;           // value returned from FindFirst/FindNext
 file_list: TStringList;    // sorted list of files in the directory
 dir_list: TStringList;     // sorted list of sub-directories
 i: integer;                // used to traverse the above lists
begin
  Application.ProcessMessages;    // allow the caller to do some work
  if not Continue then Exit;      // interrupted? - simply return to caller

  FFoundObject := dir;               // prepare to tell owner that we've
  if Assigned (FOnDirectoryFound)    // found a new directory
    then FOnDirectoryFound (self);

  file_list := TStringList.Create;   // prepare list of files
  file_list.Sorted := True;          // and note is is to be in sorted order
  dir_list := TStringList.Create;    // prepare list of directories
  dir_list.Sorted := True;           // also sorted

  dir_list.BeginUpdate;
  status := FindFirst (dir + '*.*', faAnyFile, f);  // try the first find call
  while status = 0 do with f do      // keep looping while more directories
    begin
    if ((faDirectory and Attr) <> 0) and     // is this a directory?
       (Name <> '.') and (Name <> '..')      // ignore backlinks
      then dir_list.Add (dir + Name + '\');  // add it to directory list
    status := FindNext (f);          // see if there's another directory entry
    end;
  FindClose (f);                     // all done, clean up from directory scan
  dir_list.EndUpdate;

  file_list.BeginUpdate;
  status := FindFirst (dir + FFileMask, faAnyFile, f);  // try the first find call
  while status = 0 do with f do      // keep looping while more entries
    begin
    if ((faDirectory and Attr) = 0) and   // is this not a directory?
       ((faVolumeID and Attr) = 0)        // is this not a volume ID?
      then file_list.Add (dir + Name);    // then add to file list
    status := FindNext (f);          // see if there's another directory entry
    end;
  FindClose (f);                     // all done, clean up from the scan
  file_list.EndUpdate;

  // scan the sorted list of files and call back the owner for each
  for i := 0 to file_list.Count - 1 do
    begin
    FFoundObject := file_list.Strings [i];
    if Assigned (FOnFileFound) then FOnFileFound (self);
    end;
  file_list.Free;                     // return memory owned by the list

  // scan the sorted list of directories and scan each sub-directory
  for i := 0 to dir_list.Count - 1 do
    scan_directory (dir_list.Strings [i]);
  dir_list.Free;                     // return memory owned by the list

  if Assigned (FOnDirectoryDone)     // finished this directory
    then FOnDirectoryDone (self);
end;

end.

